# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, \
    unicode_literals

__all__ = ["app", "timeline_handler"]

from flask import Flask

import chainerboard as _chainerboard


app = Flask(
    __name__,
    template_folder='templates',
    static_folder='static'
)

# FIXME: enable debug mode in beta
app.debug = True

timeline_handler = _chainerboard.TimelineHandler()

# imports on bottom to avoid import loops
import chainerboard.app.root
import chainerboard.app.events
import chainerboard.app.histograms



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import json
import logging
import time

import click


@click.command()
@click.argument('infile', type=click.Path(exists=True))
@click.argument('outfile', type=click.Path())
@click.option('-t', '--time-interval', type=float, default=10.,
              show_default=True)
@click.option('-q', '--quiet', is_flag=True, show_default=True)
def run(infile, outfile, time_interval, quiet):
    logging.basicConfig(level=logging.WARN if quiet else logging.INFO)
    logger = logging.getLogger(__name__)
    logger.info('loading input file %s ...' % infile)
    with open(infile) as fin:
        # Do not use click.File because we want close the file asap
        data = json.load(fin)
    n = len(data)
    logger.info(
        'loading input file %s done. %d data found.'% (infile, n))
    for i in xrange(len(data)):
        logger.info('Sleeping for %d sec [%d/%d] ...' % (time_interval, i+1, n))
        time.sleep(time_interval)
        with open(outfile, 'w') as fout:
            json.dump(data[:(i+1)], fout)
        logger.info('Dumped %dth/%d data to %s' % (i+1, n, outfile))


if __name__ == '__main__':
    # This script is not installed with cli.py because it is part of debugging
    # code
    run()


# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import logging

from flask import render_template, redirect, url_for

from chainerboard.app import app

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


@app.route('/')
def index():
    return redirect(url_for('events'))


@app.route('/events')
def events():
    return render_template('layouts/events.html')


@app.route('/histograms')
def histograms():
    return render_template('layouts/histograms.html')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import json
import logging
import re
import string

from chainerboard.exceptions import KeyDisappearedException, ParseError
from chainerboard.timeline import Timeline, TensorTimeline, MicroAverageTimeline

logger = logging.getLogger(__name__)


def _find_partial_match(dic, key):
    ret = []
    for k in dic.keys():
        if key in k:
            ret.append(k)
    return ret


def _find_get_patten_match_single(dic, match_func):
    """
    Find a matching key from dict. This function returns value when the
    found match is the only match in the dict.

    Args:
        dic (dict): Mapping from a key (str) to corresponding metric (numeric)
            from one output of LogReporter.
        match_func (func): A function that takes a variable (str) as input and
            returns a bool if it is a valid match.

    Returns:
        str or None

    """
    # match_func(str) -> bool
    found_key = None
    for k in dic.keys():
        if match_func(k):
            if found_key is None:
                found_key = k
            else:
                # The second match is found
                return None
    return found_key


def _find_partial_match_single(dic, key):
    return _find_get_patten_match_single(dic, lambda k, key=key: key in k)


def _find_regex_match_single(dic, pat):
    pat = re.compile(pat)
    return _find_get_patten_match_single(
        dic, lambda k, pat=pat: pat.search(k) is not None)


def _try_find_fallback(dic, key, warn=True):
    """
    Try to find a key in dic. It will try to find a exact match first. If it
    fails, it will fall back to find a partial match.

    Args:
        dic (dict): Mapping from a key (str) to corresponding metric (numeric)
            from one output of LogReporter.
        key (str): Key to find

    Returns:
        str or None

    """
    if key in dic:
        return key

    partial_matches = _find_partial_match(dic, key)
    if warn and len(partial_matches) > 1:
        logger.warn("Multiple match for key %s: %s" % (key, str(partial_matches)))
    return partial_matches[0] if len(partial_matches) > 0 else None


def _try_find_conservative(dic, key):
    """
    Try to find a key in dic. These keys are less important, so extract only if
    confident.

    These are currently dedicated to ``loss`` and ``accuracy``

    .. warning:: This function have side effects to dic

    Args:
        dic (dict): Mapping from a key (str) to corresponding metric (numeric)
            from one output of LogReporter.
        key (str): Key to find

    Returns:
        str or None: The key value
    """
    # These keys are less important, so extract only if confident
    key_default = 'main/%s' % key
    if key_default in dic:
        return key_default
    found = _find_regex_match_single(dic, r'%s$' % re.escape(key))

    return found


def match_data_grad(dic):
    """
    Search data and grad properties, which are created by ParameterStatistics.

    Args:
        dic (dict): Mapping from a key (str) to corresponding metric (numeric)
            from one output of LogReporter.

    Returns:
        dict: ret[obj_name][`data` or `grad`][metric_name] -> key
            e.g.
            ::

                {'main/link/layer/W': {
                  'data': {
                    'percentile/1': 'main/link/layer/W/data/percentile/1',
                    'zeros': 'main/link/layer/W/data/zeros',
                    ...
                    },
                  'grad': {
                    'percentile/1': 'main/link/layer/W/grad/percentile/1',
                    ...
                    }
                  },
                  ...
                }

    """
    matched = {}
    for k in dic.keys():
        if k.endswith('.std'):
            continue
        idx = string.find(k, '/data/')
        if idx >= 0:
            obj_name = k[:idx]
            if obj_name not in matched:
                matched[obj_name] = {'data': {}}
            elif 'data' not in matched[obj_name]:
                matched[obj_name]['data'] = {}
            matched[obj_name]['data'][k[idx + 6:]] = k
            continue
        idx = string.find(k, '/grad/')
        if idx < 0:
            # Not found, so continue the for loop
            continue
        obj_name = k[:idx]
        if obj_name not in matched:
            matched[obj_name] = {'grad': {}}
        elif 'grad' not in matched[obj_name]:
            matched[obj_name]['grad'] = {}
        matched[obj_name]['grad'][k[idx + 6:]] = k

    return matched


class TimelineHandler(object):
    """
    This class aggregates Timelines to handle a series of training metrics.
    """
    def __init__(self):
        self.tensors = {}
        self.events = {}
        self._time_uninitialized = True
        self._done_idx = 0

    def _initialize_time_keys(self, dic):
        self._epoch_key = _try_find_fallback(dic, 'epoch')
        self._iteration_key = _try_find_fallback(dic, 'iteration')
        self._elapsed_time_key = _try_find_fallback(dic, 'elapsed_time')
        if (self._epoch_key is None and self._iteration_key is None
                and self._elapsed_time_key is None):
            raise ParseError(
                'None of epoch, iteration and elapsed_time was found in json file.')
        self._time_uninitialized = False

    def _extract_time(self, dic):
        epoch = dic.pop(self._epoch_key, None)
        iteration = dic.pop(self._iteration_key, None)
        elapsed_time = dic.pop(self._elapsed_time_key, None)
        # Do sanity check first. Abort if something is wrong.
        if self._epoch_key is not None and epoch is None:
            raise KeyDisappearedException(self._epoch_key)
        if self._iteration_key is not None and iteration is None:
            raise KeyDisappearedException(self._iteration_key)
        if self._elapsed_time_key is not None and elapsed_time is None:
            raise KeyDisappearedException(self._elapsed_time_key)
        return epoch, iteration, elapsed_time

    def _identify_predefined_metrics(self, dic, epoch, iteration, elapsed_time):
        """
        Key name for the important metrics can change by the training configuration.
        This function determines the key names of such variables and groups
        them according its guess.

        Args:
            dic:
        """
        events = {}
        loss_key = _try_find_conservative(dic, 'loss')
        if loss_key is not None:
            events[loss_key] = Timeline(loss_key)
        accuracy_key = _try_find_conservative(dic, 'accuracy')
        if accuracy_key is not None:
            events[accuracy_key] = Timeline(accuracy_key)
        # From chainer.extensions.MicroAverage
        # These were not written in the code but is written in example
        total_key = 'main/total'
        correct_key = 'main/correct'
        if total_key in dic and correct_key in dic:
            events[correct_key] = MicroAverageTimeline(total_key, correct_key)

        for k in events.iterkeys():
            events[k].extract_value(dic, epoch, iteration, elapsed_time)
        self.events.update(events)

    def _identify_tensors(self, dic, epoch, iteration, elapsed_time):
        # For now just throw them away
        matched = match_data_grad(dic)
        tensors = {}
        for name, key_info in matched.iteritems():
            data = key_info['data']
            grad = key_info['grad']
            tensors[name] = TensorTimeline(data, grad)

        for k in tensors.iterkeys():
            tensors[k].extract_value(dic, epoch, iteration, elapsed_time)
        self.tensors.update(tensors)

    def _identify_misc(self, dic, epoch, iteration, elapsed_time):
        # For now just throw them away
        events = {}
        for k in dic.keys():
            if k.endswith('.std'):
                continue
            events[k] = Timeline(k)
        for k in events.iterkeys():
            events[k].extract_value(dic, epoch, iteration, elapsed_time)
        self.events.update(events)

    def extract(self, dic):
        if self._time_uninitialized:
            self._initialize_time_keys(dic)
        epoch, iteration, elapsed_time = self._extract_time(dic)
        for k in self.events.iterkeys():
            self.events[k].extract_value(dic, epoch, iteration, elapsed_time)
        for k in self.tensors.iterkeys():
            self.tensors[k].extract_value(dic, epoch, iteration, elapsed_time)
        self._identify_predefined_metrics(dic, epoch, iteration, elapsed_time)
        self._identify_tensors(dic, epoch, iteration, elapsed_time)
        self._identify_misc(dic, epoch, iteration, elapsed_time)
        logger.debug("Ignored keys: %s" % ', '.join(map(str, dic.keys())))

    def update(self, logs):
        """
        Update the handler from log dictionaries.
        This function is idempotent.

        ..warning :: I have not tested much that it is actually idempotent.

        Args:
            logs (list of dict): Parsed log file, each dict representing a
                report from the single output of Reporter.

        """
        while self._done_idx < len(logs):
            logger.info('Parsing line:%d' % (self._done_idx + 1))
            self.extract(logs[self._done_idx])
            self._done_idx += 1

    def get_events_ids(self):
        return self.events.keys()

    def get_tensors_ids(self):
        return self.tensors.keys()

    def load(self, path):
        """
        Aggregate records and crete timeline data that is suitable for drawing
        plots.
        """
        with open(path) as fin:
            data = json.load(fin)
        self.update(data)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, \
    unicode_literals


from chainerboard.app import events


def test_clease_plots():
    y_input = [0.1125, -0.1, float('inf'), 2.091]
    x_input = [0, 1, 2, 3]
    y = events._cleanse_plots(x_input, y_input)
    assert [0.1125, -0.1, 'nan', 2.091] == y

    # it should accept empty list
    y = events._cleanse_plots([], [])
    assert [] == y

    y_input = [float('inf')]
    x_input = [0]
    y = events._cleanse_plots(x_input, y_input)
    assert ['nan'] == y

    y_input = [0.1125, -0.1, float('nan'), 2.091]
    x_input = [0, 1, 2, 3]
    y = events._cleanse_plots(x_input, y_input)
    assert [0.1125, -0.1, 'nan', 2.091] == y

    # should not change (slightly) large value
    y_input = [1.2e5, -0.1]
    x_input = [0, 1]
    y = events._cleanse_plots(x_input, y_input)
    assert [1.2e5, -0.1] == y

    # should replace really large value
    y_input = [1.0e309, -0.1]
    x_input = [0, 1]
    y = events._cleanse_plots(x_input, y_input)
    assert ['nan', -0.1] == y



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import random
import string


_LETTERS = string.letters + string.digits


def random_hash(n, seed=None):
    random.seed(seed)
    return ''.join((random.choice(_LETTERS) for _ in xrange(n)))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import logging

import click

from chainerboard.app import app, timeline_handler
from chainerboard.watcher import watch_file

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


@click.command()
@click.argument('inputfile', type=click.Path(exists=True))
@click.option('-p', '--port', type=int, default=6006)
def cli(inputfile, port):

    with watch_file(inputfile, timeline_handler):
        # use_reloader=False because it makes watchdog unstoppable
        app.run(host='0.0.0.0', port=port, use_reloader=False)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, \
    unicode_literals


class KeyDisappearedException(Exception):
    def __init__(self, key):
        message = "Key '%s' has disappeared from data." % key
        super(KeyDisappearedException, self).__init__(message)


def warn_key_disappered(logger, key):
    logger.warn("Key '%s' has disappeared from data." % key)


class ParseError(Exception):
    pass




# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import logging

from flask import jsonify, request

from chainerboard.app import app, timeline_handler
from chainerboard import util

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


@app.route('/histograms/data', methods=['GET'])
def get_histograms_data():
    g = request.args.get('graphId')

    data = {
        'x': timeline_handler.tensors[g].iteration,
        'y': timeline_handler.tensors[g].get_percentiles(),
        'stateHash': timeline_handler.tensors[g].state_hash
    }

    return jsonify(data)


@app.route('/histograms/updates', methods=['POST'])
def get_histograms_updates():
    u"""
    Given current state, return updates.
    The `Parameters` show the content of payload, and `Returns` show the content
    of return body.

    We randomly associate groupId of 12 alphabetical + numeric characters.
    12 chars means that it needs approx 2.5e9 groups to get 0.1% chance for
    group ids to collide.

    Parameters
        active (dict): Current active graphs' states as a group id-graph ids
            pairs ``{"0d9asjan2ija": id1, ... }``
        states (dict): Mapping from graph id to state hash
        sessionId (str): Session ID that was originally given by this API.
            Should give empty string if no session ID is associated.

    Returns:
        updateType (str): 'new' or 'update'
        sessionId (str): Session ID of 12 ascii characters
        newPlots (list of dict): Newly detected plots. ``"type"`` field (str)
            tells client of action to take; ``"new"`` if the new graph does not
            belong to any group and ``"hidden"`` if it should not be shown.
            ``"graphId"`` is the graph id that client can use to get the data
            for the graph.
        updates (list of str): IDs of graphs to update (list of str).
            The list does includes graphs in newPlots.

    """
    session_id = request.json['sessionId']
    if session_id  == '':
        # no session id associated
        session_id = util.random_hash(12)
        logger.info("creating new session %s" % session_id)
        states = {}
        active = {}
        update_type = "new"
    else:
        assert isinstance(session_id, (bytes, unicode)) and len(session_id) == 12
        states = request.json['states']
        active = request.json['active']
        update_type = "update"

    # First determine graphs to add/update
    new_plots = []
    for g in timeline_handler.get_tensors_ids():
        if g not in states:
            # Currently does not implent type=append or hidden
            graph_div = util.random_hash(12)
            new_plots.append({
                'type': 'new',
                'graphDiv': graph_div,
                'graphId': g
            })
            active[graph_div] = g

    updates = []
    for group_id, g in active.iteritems():
        if timeline_handler.tensors[g].state_hash != states.get(g, ''):
            updates.append(group_id)

    return jsonify({
        'updateType': update_type,
        'sessionId': session_id,
        'newPlots': new_plots,
        'updates': updates
    })



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import contextlib
import logging
import os

from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class LogHandler(FileSystemEventHandler):
    def __init__(self, inputfile, timeline_handler):
        self._watch_path = inputfile
        self._timeline_handler = timeline_handler
        super(LogHandler, self).__init__()

    def on_modified(self, event):
        # FIXME: this is probanly plaform dependent; better way of comparing?
        if os.path.samefile(event.src_path, self._watch_path):
            logger.info("modification detected! Updating %s" % event.src_path)
            self._timeline_handler.load(event.src_path)


@contextlib.contextmanager
def watch_file(inputfile, timeline_handler):
    # log handler does not load on detecting the initial file
    timeline_handler.load(inputfile)
    event_handler = LogHandler(inputfile, timeline_handler)
    observer = Observer()
    observer.schedule(event_handler,
                      os.path.dirname(os.path.abspath(inputfile)))
    observer.start()
    yield
    observer.stop()
    observer.join()



from setuptools import setup


try:
    with open('README.rst') as f:
        readme = f.read()
except IOError:
    readme = ''

name = 'chainerboard'
exec(open('chainerboard/_version.py').read())
release = __version__
version = '.'.join(release.split('.')[:2])

setup(
    name=name,
    author='Yuta Koreeda',
    author_email='secret-email@example.com',
    maintainer='Yuta Koreeda',
    maintainer_email='secret-email@example.com',
    version=release,
    description='Visualization toolkit for Chainer',
    long_description=readme,
    url='https://github.com/koreyou/chainerboard',
    packages=[
        'chainerboard',
        'chainerboard.app'
    ],
    include_package_data=True,
    license='MIT',
    install_requires=[
        'Click',
        'flask',
        'watchdog',
        'numpy>=1.10'
    ],
    entry_points = {
        'console_scripts': ['chainerboard=chainerboard.cli:cli'],
    },
    classifiers=[
        "Environment :: Console",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Information Analysis"
    ]
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import logging
from collections import defaultdict
import warnings

from flask import jsonify, request
import numpy as np

from chainerboard import util
from chainerboard.app import app, timeline_handler

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def _cleanse_plots(x, y):
    """
    Replace NaN, inf or large number that cause harm to 'nan' (str)

    Accoding to ecma-262, largest possible value is 1.7976931348623157e308.
    Let's make anything above 1.0e308 infinity.

    Plotly can properly handle 'nan'.

    Args:
        x (list): index value of the timeline. We do not check for
            invalid value within x.
        y (list of float): Metric value to visualive.

    Returns:
        list: cleaned y

    """
    _FLOAT_CUTOFF_MAX = np.float(1.7976931348623157e308)
    _FLOAT_CUTOFF_MIN = np.float(-1.7976931348623159e308)
    x = np.asarray(x)
    y = np.asarray(y)
    assert len(x) == len(y)
    mask = np.isfinite(y)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        if np.isfinite(_FLOAT_CUTOFF_MAX):
            mask = np.logical_and(mask, y <= _FLOAT_CUTOFF_MAX)
        if np.isfinite(_FLOAT_CUTOFF_MIN):
            mask = np.logical_and(mask, y >= _FLOAT_CUTOFF_MIN)
    return [y[i] if mask[i] else 'nan' for i in xrange(len(y))]


@app.route('/events/data', methods=['GET'])
def get_events_data():
    g = request.args.get('graphId')
    logger.info(g)

    x, y = _cleanse_plots(timeline_handler.events[g].iteration,
                       timeline_handler.events[g].value)
    data = {
        'x': x,
        'y': y,
        'stateHash': timeline_handler.events[g].state_hash
    }

    return jsonify(data)


def _recursive_find_parent(graph, graph_to_group):
    if graph.id in graph_to_group:
        return graph_to_group[graph.id]
    else:
        if graph.parent is None:
            return None
        else:
            return _recursive_find_parent(graph.parent, graph_to_group)


@app.route('/events/updates', methods=['POST'])
def get_events_updates():
    u"""
    Given current state, return updates.
    The `Parameters` show the content of payload, and `Returns` show the content
    of return body.

    We randomly associate groupId of 12 alphabetical + numeric characters.
    12 chars means that it needs approx 2.5e9 groups to get 0.1% chance for
    group ids to collide.

    Parameters
        active (dict): Current active graphs' states as a group id-graph ids
            pairs ``{"0d9asjan2ija": [id1, id2, ..], ... }``
        states (dict): Mapping from graph id to state hash
        sessionId (str): Session ID that was originally given by this API.
            Should give empty string if no session ID is associated.

    Returns:
        updateType (str): 'new' or 'update'
        sessionId (str): Session ID of 12 ascii characters
        newPlots (list of dict): Newly detected plots. ``"type"`` field (str)
            tells client of action to take; ``"new"`` if the new graph does not
            belong to any group, ``"hidden"`` if it should not be shown and
            ``"append"`` if it should appended to existing group.
            ``"groupId"`` field (str) show group id to add the graph.
            ``"graphId"`` is the graph id that client can use to get the data
            for the graph.
        updates (dict): Notifies graphs to update. Maps group id (str)
            to IDs of graphs to update (list of str).
            The list does includes graphs in newPlots.

    """
    session_id = request.json['sessionId']
    if session_id  == '':
        # no session id associated
        session_id = util.random_hash(12)
        logger.info("creating new session %s" % session_id)
        states = {}
        active = {}
        update_type = "new"
    else:
        assert isinstance(session_id, (bytes, unicode)) and len(session_id) == 12
        states = request.json['states']
        active = request.json['active']
        update_type = "update"
    graph_to_group = {g: group for group, graphs in active.iteritems()
                      for g in graphs}

    # First determine graphs to add/update
    new_plots = []
    for g in timeline_handler.get_events_ids():
        if g not in states:
            # Currently does not implent type=append or hidden
            group_id = util.random_hash(12)
            new_plots.append({
                'type': 'new',
                'groupId': group_id,  # create new group id
                'name': g,
                'graphId': g
            })
            active[group_id] = [g]

    updates = defaultdict(list)
    for group_id, graphs in active.iteritems():
        for g in graphs:
            if timeline_handler.events[g].state_hash != states.get(g, ''):
                updates[group_id].append(g)

    return jsonify({
        'updateType': update_type,
        'sessionId': session_id,
        'newPlots': new_plots,
        'updates': dict(updates)
    })



# -*- coding: utf-8 -*-
"""
Deals with time-dependent data
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import logging
from collections import OrderedDict

from chainerboard.exceptions import KeyDisappearedException
from chainerboard import util

logger = logging.getLogger(__name__)


class KeyValuePair(object):
    def __init__(self, key, value):
        self.key = key
        self.value = value


class TimelineBase(object):
    def __init__(self):
        self._uninitialized = True
        self._update_state_hash()

    def _update_state_hash(self):
        self._state_hash = util.random_hash(12)

    def _initialize_times(self, epoch, iteration, time_elapsed):
        self._epoch = None if epoch is None else []
        self._iteration = None if iteration is None else []
        self._elapsed_time = None if time_elapsed is None else []
        self._uninitialized = False

    def _append_time(self, epoch, iteration, elapsed_time):
        if self._epoch is not None:
            self._epoch.append(epoch)

        if self._iteration is not None:
            self._iteration.append(iteration)

        if self._elapsed_time is not None:
            self._elapsed_time.append(elapsed_time)

    def _extract_value_impl(self, dic, epoch, iteration, elapsed_time):
        """
        Implementation for extracting of values.
        This function should return ``True`` if there were any updates otherwise
        ``False``.
        """
        raise NotImplementedError()

    def extract_value(self, dic, epoch, iteration, elapsed_time):
        if self._uninitialized:
            self._initialize_times(epoch, iteration, elapsed_time)

        ret = self._extract_value_impl(dic, epoch, iteration, elapsed_time)
        assert isinstance(ret, bool)
        if ret:
            self._update_state_hash()
        return ret

    @property
    def elapsed_time(self):
        return self._elapsed_time

    @property
    def epoch(self):
        return self._epoch

    @property
    def iteration(self):
        return self._iteration

    @property
    def state_hash(self):
        return self._state_hash


class Timeline(TimelineBase):
    def __init__(self, key):
        super(Timeline, self).__init__()
        self._key = key
        self._key_std = key + '.std'
        self._value = []

    def _extract_value_impl(self, dic, epoch, iteration, elapsed_time):
        updated = False
        if self._key_std in dic:
            std = dic.pop(self._key_std)
            updated = True
        if self._key in dic:
            val = dic.pop(self._key)
            self._value.append(val)
            self._append_time(epoch, iteration, elapsed_time)
            updated = True
        return updated

    @property
    def value(self):
        return self._value


class MicroAverageTimeline(TimelineBase):
    def __init__(self, total_key, correct_key):
        super(MicroAverageTimeline, self).__init__()
        self._total_key = total_key
        self._correct_key = correct_key
        self._value = []

    def _extract_value_impl(self, dic, epoch, iteration, elapsed_time):
        updated = False
        if self._correct_key in dic:
            # Assume that both keys are in dic
            correct = dic.pop(self._correct_key)
            try:
                total = dic.pop(self._total_key)
            except KeyError:
                raise KeyDisappearedException(self._total_key)
            micro_average = correct / float(total)
            self._value.append(micro_average)
            self._append_time(epoch, iteration, elapsed_time)
            updated = False
        elif self._total_key in dic:
            # Only total_key exists in dic -> raise Error
            raise KeyDisappearedException(self._correct_key)
        return updated

    @property
    def value(self):
        return self._value


class TensorTimeline(TimelineBase):
    """
    Timeline for visualizing tensors. This is intended for logs from
    [chainer.extensions.ParameterStatistics](https://docs.chainer.org/en/latest/reference/generated/chainer.training.extensions.ParameterStatistics.html)

    .. warning:: Class constructor has side effect to ``data_keys`` and
                 ``grad_keys``

    Args:
        data_keys (dict): Mapping from metric name (str) to key string (str).
        grad_keys (dict): Mapping from metric name (str) to key string (str).

    """

    _DEFAULT_PERCENTILE_KEYS = {
        'percentile/0': r'0.13%',
        'percentile/1': r'2.28%',
        'percentile/2': r'15.87%',
        'percentile/3': r'50%',
        'percentile/4': r'84.13%',
        'percentile/5': r'97.72%',
        'percentile/6': r'99.87%',
    }
    def __init__(self, data_keys, grad_keys):
        super(TensorTimeline, self).__init__()
        self._data_percentiles_keys = {
            k: data_keys.pop(k) for k in data_keys.keys() if k.startswith('percentile/')}
        try:
            self._data_percentiles_names = {
                k: self._DEFAULT_PERCENTILE_KEYS[k]
                for k in self._data_percentiles_keys.keys()
            }
        except KeyError:
            self._data_percentiles_names = {
                k: k for k in self._data_percentiles_keys.keys()
            }
        self._data_percentiles = OrderedDict(
            ((k, []) for k in sorted(self._data_percentiles_keys.keys()))
        )
        self._data = {k: [] for k in data_keys.keys()}
        self._data_keys = data_keys

        self._grad_percentiles_keys = {
            k: grad_keys.pop(k) for k in grad_keys.keys() if k.startswith('percentile/')}
        self._grad_percentiles = OrderedDict(
            ((k, []) for k in sorted(self._grad_percentiles_keys.keys()))
        )
        self._grad = {k: [] for k in grad_keys.keys()}
        self._grad_keys = grad_keys

    def _extract_value_impl(self, dic, epoch, iteration, elapsed_time):
        if next(self._data_percentiles_keys.itervalues()) not in dic:
            # Assume tensor is absent in dic
            return False
        self._append_time(epoch, iteration, elapsed_time)
        try:
            for k, dic_key in self._data_percentiles_keys.iteritems():
                self._data_percentiles[k].append(dic.pop(dic_key))
            for k, dic_key in self._data_keys.iteritems():
                self._data[k].append(dic.pop(dic_key))
            for k, dic_key in self._grad_percentiles_keys.iteritems():
                self._grad_percentiles[k].append(dic.pop(dic_key))
            for k, dic_key in self._grad_keys.iteritems():
                self._grad[k].append(dic.pop(dic_key))
        except KeyError:
            raise KeyDisappearedException(dic_key)
        return True

    def get_percentiles(self):
        percentiles = [
            {'label': self._data_percentiles_names[k], 'data': v}
            for k, v in self._data_percentiles.iteritems()
        ]
        if 'min' in self._data and 'max' in self._data:
            percentiles = (
                [{'label': 'min', 'data': self._data['min']}, ] +
                percentiles +
                [{'label': 'max', 'data': self._data['max']}, ]
            )
        return percentiles
